import h5py
import numpy as np
from numpy import sqrt, pi
from qosm_core import Vec3

from qosm.sources.mode_matching.mm_run import rect_aperture_modes
from qosm.utils.Field import Field
from qosm.utils.Pose import Vector, Frame, Quaternion
from qosm.sources.Source import Source


def trapz2d(M, dx, dy):
    N = M.shape[1]
    int_y = dx * ((M[:, 0:N - 1] - M[:, 1:N]) / 2. + M[:, 1:N])
    Y = np.sum(int_y, axis=1)
    int_x = dy * ((Y[0:N - 1] - Y[1:N]) / 2. + Y[1:N])
    return np.sum(int_x)


class Horn(Source):
    TYPE = '_HORN_'

    def __init__(self, frequency_GHz: float):
        super(Horn, self).__init__(frequency_GHz)
        self.Lx = 0
        self.Ly = 0
        self.Lz = 0
        self.Ea = np.array([], dtype=complex)
        self.Ha = np.array([], dtype=complex)
        self.xa_ = np.array([])
        self.ya_ = np.array([])
        self.Nax = 0
        self.Nay = 0
        self.rot_z_deg = 0

    def init(self, **kwargs):
        if 'rot_z_deg' in kwargs:
            self.rot_z_deg = kwargs['rot_z_deg']

        if 'Lx' in kwargs and 'Ly' in kwargs and 'modes' in kwargs and 'num_pts_aperture' in kwargs:
            xa = np.linspace(-kwargs['Lx'] / 2, kwargs['Lx'] / 2, kwargs['num_pts_aperture'], endpoint=True)
            ya = np.linspace(-kwargs['Ly'] / 2, kwargs['Ly'] / 2, kwargs['num_pts_aperture'], endpoint=True)

            ck = kwargs['modes']
            Nm = int(ck.shape[1])
            Ek, Hk, Zk, mode_names = rect_aperture_modes(self.freq, Nm, xa, ya)

            Ek *= ck.reshape((1, 1, -1, 1))
            Hk *= ck.reshape((1, 1, -1, 1))
            Ea = np.sum(Ek, axis=2)
            Ha = np.sum(Hk, axis=2)

            Ea = Ea.reshape((-1, 3))
            Ha = Ha.reshape((-1, 3))

            # call the method again to set the aperture fields
            self.init(fields=(Ea, Ha), axis=(xa, ya))

        if 'fields' in kwargs and 'axis' in kwargs:
            self.Ea, self.Ha = kwargs['fields']
            self.xa_, self.ya_ = kwargs['axis']

            self.xa_ = np.reshape(self.xa_, (-1,))
            self.ya_ = np.reshape(self.ya_, (-1,))
            self.Lx = np.max(self.xa_) - np.min(self.xa_)
            self.Ly = np.max(self.ya_) - np.min(self.ya_)
            self.Nax = self.xa_.shape[0]
            self.Nay = self.ya_.shape[0]
            self.Ea = np.reshape(self.Ea, (self.Nay, self.Nax, 3))
            self.Ha = np.reshape(self.Ha, (self.Nay, self.Nax, 3))

        elif 'h5' in kwargs:
            filename = kwargs['h5']
            with h5py.File(filename, "r") as f:
                self.Ea = f['Eapt_glo']['real'][()] + f['Eapt_glo']['imag'][()] * 1j
                self.Ha = f['Hapt_glo']['real'][()] + f['Hapt_glo']['imag'][()] * 1j
                self.xa_ = np.reshape(f['xa_'][()], (-1,))
                self.ya_ = np.reshape(f['ya_'][()], (-1,))
                self.Lx = np.max(self.xa_) - np.min(self.xa_)
                self.Ly = np.max(self.ya_) - np.min(self.ya_)

                self.Ea = np.reshape(self.Ea, (3, -1)).T
                self.Ha = np.reshape(self.Ha, (3, -1)).T
                self.Nax = self.xa_.shape[0]
                self.Nay = self.ya_.shape[0]
                self.Ea = np.reshape(self.Ea, (self.Nay, self.Nax, 3))
                self.Ha = np.reshape(self.Ha, (self.Nay, self.Nax, 3))

    def aperture_points(self, n_samples_per_axis: int) -> (Vector, np.ndarray, np.ndarray):
        xa_ = np.linspace(-self.Lx / 2, self.Lx / 2, n_samples_per_axis)
        ya_ = np.linspace(-self.Ly / 2, self.Ly / 2, n_samples_per_axis)
        qrotz = Quaternion(axis=Vector(0, 0, 1), angle=self.rot_z_deg, deg=True)
        Xa_, Ya_ = np.meshgrid(xa_, ya_)
        r_ap_diag0 = Vector(N=xa_.shape[0] * ya_.shape[0])
        r_ap_diag0[:, 0] = np.reshape(Xa_, (-1,))
        r_ap_diag0[:, 1] = np.reshape(Ya_, (-1,))
        r_ap_diag0[:, 2] = 0.0
        r_ap_src = self.frame.to_ref(qrotz.rotate(r_ap_diag0))
        return r_ap_src, xa_, ya_

    def e_field(self, r_pts_src: Vector, pbar=None):
        # apply rotation around src z-axis: example -> case of diagonal horn
        rot_z = Frame(quaternion=Quaternion(axis=Vector(0, 0, 1), angle=self.rot_z_deg, deg=True))
        r_pts_src2 = rot_z.rot_to_ref(r_pts_src)

        N = r_pts_src.shape[0]
        Ni = self.xa_.shape[0]
        n, kappa = (1., 0.)
        k = self.k0 * (n - kappa * 1j)

        # Electric field Output initialisation
        E = Field(N=N)

        (Xa, Ya) = np.meshgrid(self.xa_, self.ya_)
        if pbar is not None:
            pbar.emit(0, N)

        dx = np.abs(self.xa_[1] - self.xa_[0])
        dy = np.abs(self.ya_[1] - self.ya_[0])
        dS = dx * dy

        Z = 2 * sqrt(pi * 1e-7 / 8.854187817e-12) / (n - kappa * 1j)
        inv_Z = 1. / Z

        # for each point
        for ir in range(0, N):
            if pbar is not None:
                pbar.emit(ir + 1, 0)
            r_pts = r_pts_src2[ir, :]
            r_hat = np.zeros((self.xa_.size, self.ya_.size, 3), dtype=np.clongdouble)

            r_hat[:, :, 0] = r_pts[0] - Xa
            r_hat[:, :, 1] = r_pts[1] - Ya
            r_hat[:, :, 2] = r_pts[2]
            r = np.sqrt(r_hat[:, :, 0] ** 2 + r_hat[:, :, 1] ** 2 + r_hat[:, :, 2] ** 2).reshape((Ni, Ni, 1))
            r_hat = r_hat / r

            # Electric surface currents
            J = np.zeros((Ni, Ni, 3), dtype=complex)
            J[:, :, 0] = - self.Ha[:, :, 1]
            J[:, :, 1] = self.Ha[:, :, 0]

            # Magnetic surface currents
            M = np.zeros((Ni, Ni, 3), dtype=complex)
            M[:, :, 0] = self.Ea[:, :, 1]
            M[:, :, 1] = - self.Ea[:, :, 0]

            M_cross_r = np.zeros((Ni, Ni, 3), dtype=complex)
            M_cross_r[:, :, 0] = M[:, :, 1] * r_hat[:, :, 2] - M[:, :, 2] * r_hat[:, :, 1]
            M_cross_r[:, :, 1] = M[:, :, 2] * r_hat[:, :, 0] - M[:, :, 0] * r_hat[:, :, 2]
            M_cross_r[:, :, 2] = M[:, :, 0] * r_hat[:, :, 1] - M[:, :, 1] * r_hat[:, :, 0]

            r_dot_J = np.sum(r_hat * J.conj(), axis=2).reshape((Ni, Ni, 1))

            inv_kr = 1. / (k * r)

            # Green Function
            G = np.exp(-1j * k * r) * inv_kr

            A = 1 - 1j * inv_kr - inv_kr ** 2
            B = - 1 + 3j * inv_kr - 3 * inv_kr ** 2
            C = 1 - 1j * inv_kr
            dE = (A * J + B * r_dot_J * r_hat + inv_Z * C * M_cross_r) * G

            # dE integration along x and y
            mE = -1j * Z * k * k / (4 * np.pi) * np.sum(np.sum(dE, axis=1), axis=0) * dS

            E[ir, :] = mE
        E = rot_z.rot_to_local(E)

        return E

    def h_field(self, r_pts_src: Vector, pbar=None):
        # apply rotation around src z-axis: example -> case of diagonal horn
        rot_z = Frame(quaternion=Quaternion(axis=Vector(0, 0, 1), angle=self.rot_z_deg, deg=True))
        r_pts_src2 = rot_z.rot_to_ref(r_pts_src)

        N = r_pts_src.shape[0]
        Ni = self.xa_.shape[0]
        n, kappa = (1., 0.)
        k = self.k0 * (n - kappa * 1j)

        # Magnetic field Output initialisation
        H = Field(N=N)

        (Xa, Ya) = np.meshgrid(self.xa_, self.ya_)
        if pbar is not None:
            pbar.emit(0, N)

        dx = np.abs(self.xa_[1] - self.xa_[0])
        dy = np.abs(self.ya_[1] - self.ya_[0])
        dS = dx * dy

        Z = 2 * sqrt(pi * 1e-7 / 8.854187817e-12) / (n - kappa * 1j)

        # for each point
        for ir in range(0, N):
            if pbar is not None:
                pbar.emit(ir + 1, 0)
            r_pts = r_pts_src2[ir, :]
            r_hat = np.zeros((self.xa_.size, self.ya_.size, 3), dtype=np.clongdouble)

            r_hat[:, :, 0] = Xa - r_pts[0]
            r_hat[:, :, 1] = Ya - r_pts[1]
            r_hat[:, :, 2] = - r_pts[2]
            r = np.sqrt(r_hat[:, :, 0] ** 2 + r_hat[:, :, 1] ** 2 + r_hat[:, :, 2] ** 2).reshape((Ni, Ni, 1))
            r_hat = r_hat / r

            # Electric surface currents
            J = np.zeros((Ni, Ni, 3), dtype=complex)
            J[:, :, 0] = - self.Ha[:, :, 1]
            J[:, :, 1] = self.Ha[:, :, 0]

            # Magnetic surface currents
            M = np.zeros((Ni, Ni, 3), dtype=complex)
            M[:, :, 0] = - self.Ea[:, :, 1]
            M[:, :, 1] = self.Ea[:, :, 0]

            J_cross_r = np.zeros((Ni, Ni, 3), dtype=complex)
            J_cross_r[:, :, 0] = J[:, :, 1] * r_hat[:, :, 2] - J[:, :, 2] * r_hat[:, :, 1]
            J_cross_r[:, :, 1] = J[:, :, 2] * r_hat[:, :, 0] - J[:, :, 0] * r_hat[:, :, 2]
            J_cross_r[:, :, 2] = J[:, :, 0] * r_hat[:, :, 1] - J[:, :, 1] * r_hat[:, :, 0]

            r_dot_M = np.sum(r_hat * M.conj(), axis=2).reshape((Ni, Ni, 1))

            inv_kr = 1. / (k * r)

            # Green Function
            G = - np.exp(-1j * k * r) * inv_kr

            A = 1 - 1j * inv_kr - inv_kr ** 2
            B = - 1 + 3j * inv_kr - 3 * inv_kr ** 2
            C = 1 - 1j * inv_kr
            dH = (A * M + B * r_dot_M * r_hat - Z * C * J_cross_r) * G

            # dH integration along x and y
            mH = 1j * k * k / (4 * Z * np.pi) * np.sum(np.sum(dH, axis=1), axis=0) * dS

            H[ir, :] = mH

        H = rot_z.rot_to_local(H)
        return H

    def fields(self, r_pts_src: Vector, pbar=None):
        # apply rotation around src z-axis: example -> case of diagonal horn
        rot_z = Frame(quaternion=Quaternion(axis=Vector(0, 0, 1), angle=self.rot_z_deg, deg=True))
        r_pts_src2 = rot_z.rot_to_local(r_pts_src)

        N = r_pts_src.shape[0]
        k = self.k0
        Z = 2 * sqrt(pi * 1e-7 / 8.854187817e-12)  # / (n - kappa * 1j)
        inv_Z = 1. / Z

        # Electric field Output initialisation
        E = Field(N=N)

        # Magnetic field Output initialisation
        H = Field(N=N)

        # Poynting Vector Output initialisation
        P = Vector(N=N)

        (Xa, Ya) = np.meshgrid(self.xa_, self.ya_)
        Xa = Xa.flatten()
        Ya = Ya.flatten()
        Nap = Xa.shape[0]
        if pbar is not None:
            pbar.emit(0, N)

        dx = np.abs(self.xa_[1] - self.xa_[0])
        dy = np.abs(self.ya_[1] - self.ya_[0])
        dS = dx * dy

        cte1 = -1j * Z * k * k / (4 * np.pi) * dS
        cte2 = 1j * k * k / (4 * Z * np.pi) * dS

        # for each point
        for ir in range(0, N):
            if pbar is not None:
                pbar.emit(ir + 1, 0)
            r_pts = r_pts_src2[ir, :]
            r_hat = Vector(N=Nap)

            r_hat[:, 0] = Xa - r_pts[0]
            r_hat[:, 1] = Ya - r_pts[1]
            r_hat[:, 2] = - r_pts[2]
            r = r_hat.norm().reshape((-1, 1))
            r_hat = r_hat / r

            # Electric surface currents
            J = Field(N=Nap)
            J[:, 0] = - self.Ha.reshape((-1, 3))[:, 1]
            J[:, 1] = self.Ha.reshape((-1, 3))[:, 0]

            # Magnetic surface currents
            M = Field(N=Nap)
            M[:, 0] = - self.Ea.reshape((-1, 3))[:, 1]
            M[:, 1] = self.Ea.reshape((-1, 3))[:, 0]

            r_cross_M = Field(N=Nap)
            r_cross_M[:, 0] = - r_hat[:, 2] * M[:, 1]
            r_cross_M[:, 1] = r_hat[:, 2] * M[:, 0]
            r_cross_M[:, 2] = r_hat[:, 0] * M[:, 1] - r_hat[:, 1] * M[:, 0]

            r_cross_J = Field(N=Nap)
            r_cross_J[:, 0] = - r_hat[:, 2] * J[:, 1]
            r_cross_J[:, 1] = r_hat[:, 2] * J[:, 0]
            r_cross_J[:, 2] = r_hat[:, 0] * J[:, 1] - r_hat[:, 1] * J[:, 0]

            r_dot_J = np.sum(r_hat * J.conj(), axis=1).reshape((Nap, 1))
            r_dot_M = np.sum(r_hat * M.conj(), axis=1).reshape((Nap, 1))
            inv_kr = 1. / (k * r)

            # Green Function
            G = np.exp(-1j * k * r) * inv_kr

            A = 1 - 1j * inv_kr - inv_kr ** 2
            B = - 1 + 3j * inv_kr - 3 * inv_kr ** 2
            C = 1 - 1j * inv_kr
            dE = (A * J + B * r_dot_J * r_hat + inv_Z * C * r_cross_M) * G
            dH = (A * M + B * r_dot_M * r_hat - Z * C * r_cross_J) * G

            # dE integration along x and y
            mE = cte1 * np.sum(dE, axis=0)
            mH = cte2 * np.sum(dH, axis=0)

            E[ir, :] = mE
            H[ir, :] = mH

            # Compute Poynting vector
            mHc = np.conj(mH)
            P[ir, :] = 0.5 * np.real([
                mE[1] * mHc[2] - mE[2] * mHc[1],
                mE[2] * mHc[0] - mE[0] * mHc[2],
                mE[0] * mHc[1] - mE[1] * mHc[0]
            ])
        E = rot_z.rot_to_ref(E)
        H = rot_z.rot_to_ref(H)
        P = rot_z.rot_to_ref(P)

        return E, H, P

    def oriented_fields(self, r_pts_src: Vector, z_probe_src: Vec3, pbar=None):
        # apply rotation around src z-axis: example -> case of diagonal horn
        rot_z = Frame(quaternion=Quaternion(axis=Vector(0, 0, 1), angle=self.rot_z_deg, deg=True))
        r_pts_src2 = rot_z.rot_to_local(r_pts_src)  # why to ref initialy and then to local ???

        N = r_pts_src.shape[0]
        Ni = self.xa_.shape[0]
        n, kappa = (1., 0.)
        k = self.k0 * (n - kappa * 1j)
        Z = 2 * sqrt(pi * 1e-7 / 8.854187817e-12) / (n - kappa * 1j)
        inv_Z = 1 / Z

        # Electric field Output initialisation
        E = Field(N=N)

        # Magnetic field Output initialisation
        H = Field(N=N)

        # Poynting Output initialisation
        P = Vector(N=N)

        (Xa, Ya) = np.meshgrid(self.xa_, self.ya_)
        if pbar is not None:
            pbar.emit(0, N)

        dx = np.abs(self.xa_[1] - self.xa_[0])
        dy = np.abs(self.ya_[1] - self.ya_[0])
        dS = dx * dy

        # for each point
        for ir in range(0, N):
            if pbar is not None:
                pbar.emit(ir + 1, 0)
            r_pts = r_pts_src2[ir, :]
            r_hat = np.zeros((self.xa_.size, self.ya_.size, 3), dtype=np.clongdouble)

            r_hat[:, :, 0] = Xa - r_pts[0]
            r_hat[:, :, 1] = Ya - r_pts[1]
            r_hat[:, :, 2] = - r_pts[2]
            r = np.sqrt(r_hat[:, :, 0] ** 2 + r_hat[:, :, 1] ** 2 + r_hat[:, :, 2] ** 2).reshape((Ni, Ni, 1))
            r_hat = r_hat / r

            # Electric surface currents
            J = np.zeros((Ni, Ni, 3), dtype=complex)
            J[:, :, 0] = - self.Ha[:, :, 1]
            J[:, :, 1] = self.Ha[:, :, 0]

            # Magnetic surface currents
            M = np.zeros((Ni, Ni, 3), dtype=complex)
            M[:, :, 0] = - self.Ea[:, :, 1]
            M[:, :, 1] = self.Ea[:, :, 0]

            r_cross_M = np.zeros((Ni, Ni, 3), dtype=complex)
            r_cross_M[:, :, 0] = r_hat[:, :, 1] * M[:, :, 2] - r_hat[:, :, 2] * M[:, :, 1]
            r_cross_M[:, :, 1] = r_hat[:, :, 2] * M[:, :, 0] - r_hat[:, :, 0] * M[:, :, 2]
            r_cross_M[:, :, 2] = r_hat[:, :, 0] * M[:, :, 1] - r_hat[:, :, 1] * M[:, :, 0]

            r_cross_J = np.zeros((Ni, Ni, 3), dtype=complex)
            r_cross_J[:, :, 0] = r_hat[:, :, 1] * J[:, :, 2] - r_hat[:, :, 2] * J[:, :, 1]
            r_cross_J[:, :, 1] = r_hat[:, :, 2] * J[:, :, 0] - r_hat[:, :, 0] * J[:, :, 2]
            r_cross_J[:, :, 2] = r_hat[:, :, 0] * J[:, :, 1] - r_hat[:, :, 1] * J[:, :, 0]

            r_dot_J = np.sum(r_hat * J.conj(), axis=2).reshape((Ni, Ni, 1))
            r_dot_M = np.sum(r_hat * M.conj(), axis=2).reshape((Ni, Ni, 1))

            # Green Function
            G = - np.exp(-1j * k * r) / r

            inv_kr = 1 / (k * r)
            A = 1 - 1j * inv_kr - inv_kr ** 2
            B = - 1 + 3j * inv_kr + 3 * inv_kr ** 2
            C = 1 - 1j * inv_kr
            dE = (A * J + B * r_dot_J * r_hat - inv_Z * C * r_cross_M) * G
            dH = (A * M + B * r_dot_M * r_hat + Z * C * r_cross_J) * G

            # dE integration along x and y
            mE = -1j * Z * k / (4 * np.pi) * np.sum(np.sum(dE, axis=1), axis=0) * dS
            mH = 1j * k / (4 * Z * np.pi) * np.sum(np.sum(dH, axis=1), axis=0) * dS

            E[ir, :] = mE
            H[ir, :] = mH

            # Compute Poynting vector
            mHc = np.conj(mH)
            P[ir, :] = 0.5 * np.real([
                mE[1] * mHc[2] - mE[2] * mHc[1],
                mE[2] * mHc[0] - mE[0] * mHc[2],
                mE[0] * mHc[1] - mE[1] * mHc[0]
            ])
        E = rot_z.rot_to_ref(E)
        H = rot_z.rot_to_ref(H)
        P = rot_z.rot_to_ref(P)

        if z_probe_src is not None:
            s = P[:, 0]*z_probe_src[0] + P[:, 1]*z_probe_src[1] + P[:, 2]*z_probe_src[2]
            E[s < 0] = 0.
            H[s < 0] = 0.
            P[s < 0] = 0.

        return E, H, P

    def to_local_frame(self, r_a_glo: Vector, q_a_glo=None, z_rot=True):
        r_a_src = self.frame.to_local(r_a_glo)
        return r_a_src
