import numpy as np

from qosm.sources.Source import Source
from qosm.utils.Field import Field
from qosm.utils.Pose import Vector


class VirtualSource(Source):
    TYPE = '_VSRC_'

    def __init__(self, frequency_GHz: float):
        super(VirtualSource, self).__init__(frequency_GHz)
        self._chunk_size = 50
        self._num_tiles = 1
        self._proc_ratio = 0.8
        self._beams = []

    @property
    def beams(self) -> list:
        return self._beams

    def init(self, **kwargs):
        if 'beams' in kwargs:
            self._beams = kwargs['beams']

        if 'chunk_size' in kwargs:
            self._chunk_size = kwargs['chunk_size']
        if 'num_tiles' in kwargs:
            self._num_tiles = kwargs['num_tiles']
        if 'proc_ratio' in kwargs:
            self._proc_ratio = kwargs['proc_ratio']

    def e_field(self, r_pts_src, pbar=None) -> Field:
        E_src = Field(N=r_pts_src.shape[0])

        if pbar is not None:
            pbar.emit(0, len(self._beams))

        n = 0
        for ray in self._beams:
            E_, _ = ray.compute_fields(r_pts_src, )
            E_src += E_
            if pbar is not None:
                pbar.emit(n)
            n += 1

        E_src[np.isnan(E_src)] = 0.

        return E_src

    def h_field(self, r_pts_src, pbar=None) -> Field:
        H_src = Field(N=r_pts_src.shape[0])

        if pbar is not None:
            pbar.emit(0, len(self._beams))

        n = 0
        for ray in self._beams:
            _, H_ = ray.compute_fields(r_pts_src, )
            H_src += H_
            if pbar is not None:
                pbar.emit(n)
            n += 1

        H_src[np.isnan(H_src)] = 0.

        return H_src

    def fields(self, r_pts_src, pbar=None) -> (Field, Field, Vector):

        if pbar is not None:
            pbar.emit(0, len(self._beams))

        n = 0

        E_src = Field(N=r_pts_src.shape[0])
        H_src = Field(N=r_pts_src.shape[0])
        for ray in self._beams:
            E_, H_ = ray.compute_fields(r_pts_src, )
            E_src += E_
            H_src += H_
            if pbar is not None and len(self._beams) > 1:
                pbar.emit(n)
            n += 1

        # Poynting vector
        P = Vector(N=E_src.shape[0])
        P[:, 0] = 0.5 * np.real(E_src[:, 1] * H_src[:, 2] - E_src[:, 2] * H_src[:, 1])
        P[:, 1] = 0.5 * np.real(E_src[:, 2] * H_src[:, 0] - E_src[:, 0] * H_src[:, 2])
        P[:, 2] = 0.5 * np.real(E_src[:, 0] * H_src[:, 1] - E_src[:, 1] * H_src[:, 0])

        return E_src, H_src, P

    def reset(self) -> None:
        self._beams = []

    def add_beams(self, beams) -> None:
        self._beams.extend(beams)

    def get_nb_rays(self) -> int:
        return len(self._beams)
