import numpy as np
from qosm_core import VirtualSource, Vec3
from scipy import integrate

from qosm.sources.GaussianBeam import GaussianBeam
from qosm.sources.Source import Source
from qosm.sources.mode_matching.mm_run import rect_aperture_modes
from qosm.utils.Pose import Frame, Quaternion, Vector, MeshGrid


def create_gaussian_beam(
        freq_GHz: float = 300.,
        w0_mm: float = 2.3,
        z0_mm: float = 0.,
        pol_xz: float = 0.,
        pol_yz: float = 1.
) -> GaussianBeam:
    w0 = w0_mm * 1e-3
    z0 = z0_mm * 1e-3
    src = GaussianBeam(freq_GHz)
    src.init(w0=w0, polarisation=(pol_xz, pol_yz), z0=z0)
    return src


def create_virtual_source(
        freq_GHz: float = 300.,
        beams: list = ()
) -> VirtualSource:
    vsrc = VirtualSource(freq_GHz)
    vsrc.init(beams=beams)
    return vsrc


def compute_rx_power(
        src,
        Lx: float,
        Ly: float,
        Npts: int,
        Nm: int,
        rot_z_deg: float,
        pts: Vector,
        sens: int = 0,
        pbar=None
):
    freq = src.freq
    rot_z = Frame(quaternion=Quaternion(axis=Vector(0, 0, 1), angle=rot_z_deg, deg=True))

    xa = np.linspace(-Lx / 2, Lx / 2, Npts, endpoint=True)
    ya = np.linspace(-Ly / 2, Ly / 2, Npts, endpoint=True)

    Ek, Hk, Zk, mode_names = rect_aperture_modes(freq, Nm, xa, ya)
    Ek *= -1

    Pmz = Ek[:, :, :, 0] * Hk[:, :, :, 1].conj() - Ek[:, :, :, 1] * Hk[:, :, :, 0].conj()
    Pk = integrate.trapezoid(integrate.trapezoid(Pmz, ya, axis=1), xa, axis=0)

    Ekc = Ek.conj()
    Hkc = Hk.conj()
    Yk = 1 / Zk

    # construct array for each points in witch to measure
    # SO FAR, we do not allow for surface tilt => only on xy plane
    g = MeshGrid(u=xa, v=ya)
    pts_xy = g.xy_plane(z=0)
    pts_xy = rot_z.rot_to_local(pts_xy)

    P = np.zeros((pts.shape[0],), dtype=complex)
    ck = np.zeros((pts.shape[0], Pk.shape[0]), dtype=complex)

    pts_meas = np.zeros((pts.shape[0], Npts * Npts, 3))
    for i in range(pts.shape[0]):
        pts_meas[i, :, :] = pts_xy + pts[i, :]

    pts_meas = pts_meas.reshape((-1, 3))

    if sens == 0:
        probe_z_src = Vec3(0., 0., 1.)
    else:
        probe_z_src = Vec3(0., 0., -1.)

    try:
        E, _, _ = src.oriented_fields(r_pts_src=pts_meas, z_probe_src=probe_z_src, pbar=pbar)
    except Exception as _:
        E, _, _ = src.oriented_fields(r_pts_src=pts_meas, z_probe_src=probe_z_src)
        E = E.numpy()

    if sens == 1:
        rot_y = Frame(quaternion=Quaternion(axis=Vector(0, 1, 0), angle=180, deg=True))
        E = rot_y.rot_to_ref(E)

    E = rot_z.rot_to_ref(E)

    E = E.reshape((pts.shape[0], Npts * Npts, 3))

    if pbar is not None:
        pbar.emit(0, pts.shape[0])
    for i in range(pts.shape[0]):
        E_field_sim = E[i, :].reshape((Npts, Npts, 1, 3))

        cross_dot_z = Yk * np.sum(E_field_sim.reshape((Npts, Npts, 1, 3)) * Ekc, axis=3)
        modes_ck = integrate.trapezoid(integrate.trapezoid(cross_dot_z, ya, axis=1), xa, axis=0)
        P[i] = np.sum(modes_ck ** 2)
        ck[i, :] = modes_ck
        if pbar is not None:
            pbar.emit(i + 1, 0)

    return P, ck


def oewg_probe_s21(src: Source, probe_poses: np.ndarray, Lx_probe=0.8636e-3, Ly_probe=0.4318e-3, Npts=51):
    r_probe_src = Vector(probe_poses[:, 0:3])
    f_probe_src = Frame(pose=probe_poses)
    E_src = src.e_field(r_probe_src)

    E_probe = f_probe_src.rot_to_local(E_src)

    xa_probe = np.linspace(-Lx_probe / 2, Lx_probe / 2, Npts, endpoint=True)
    ya_probe = np.linspace(-Ly_probe / 2, Ly_probe / 2, Npts, endpoint=True)
    Ek_probe, Hk_probe, Zk_probe, _ = rect_aperture_modes(src.freq, 2, xa_probe, ya_probe)
    Yk_probe = 1 / Zk_probe
    Ekc_probe = np.conj(Ek_probe)

    S21 = np.zeros((E_src.shape[0],), dtype=complex)
    for i in range(E_src.shape[0]):
        # consider the field uniform on the aperture
        E_field_sim = np.ones((Npts, Npts, 3)) * np.reshape(E_src[i, :], (1, 1, 3))
        cross_dot_z = np.sum(E_field_sim.reshape((Npts, Npts, 1, 3)) * Ekc_probe, axis=3)
        modes_ck = 2.**.5 * Yk_probe * integrate.trapezoid(
            integrate.trapezoid(cross_dot_z, ya_probe, axis=1), xa_probe, axis=0
        )
        S21[i] = modes_ck[0]

    return S21
